################################################################
#            The mmsearch Metamotif Search Engine
#
# Thomas Junier, April 2001
#
# Please see file COPYRIGHT.
#
# mmParse.py : a metamotif parser for generating an FSA (and then some
# :-).

# The `input' is a list of (value, type) tuples representing the tokens,
# as returned by mmScan. Parameter `stack' is a state stack used by the
# automaton. The parser needs it because it directly builds the
# automaton, rather than generating a parse tree. The set of mutually
# called functions implement a recursive-descent parser. The `level'
# parameter indicates the recursion level, it is used by each function
# when it prints that it's being called, resulting in an approximation
# of the parse tree - this useful mainly for debugging, it might get
# suppressed when everything appears to work fine.

# All functions defined perform two tasks (that may be intertwined):
# they parse the input tokens into whichever structure they are named
# after (or raise a SyntaxError if they can't); and they generate
# automata either by creating new ones or by linking automata into
# larger ones. The top-level function, metamotif(), generates the
# whole automaton by assembling the automata generated by the functions
# it called, etc.

# Here is the grammar:

# METAMOTIF     ::= (SEQUENCE|FORK)+ 
# SEQUENCE      ::= (FEATURE_LIST|GROUP)+
# FORK          ::= L_BRACKET BANG? FEATURE_LIST {PIPE FEATURE_LIST}* R_BRACKET
# GROUP         ::= L_PAREN SEQUENCE {PIPE SEQUENCE}* R_PAREN {RANGE}
# FEATURE_LIST  ::= FEATURE_BLOCK {SPACER FEATURE_BLOCK}*
# FEATURE_BLOCK ::= FEATURE_END+
# SPACER        ::= INTEGER COMMA INTEGER
# |             ::= COMMA INTEGER
# |             ::= INTEGER COMMA
# FEATURE_END   ::= FEATURE_START | FEATURE_STOP
# FEATURE_START ::= L_A_BRACKET LETTER  {HASH (LETTER | DIGIT)}?
# FEATURE_STOP  ::= LETTER R_A_BRACKET {HASH (LETTER | DIGIT)}?
# INTEGER       ::= DIGIT+
# LETTER        ::= ['A'-'Z'a'-'z']
# DIGIT         ::= ['0'-'9']
# L_PAREN       ::= '('
# R_PAREN       ::= ')'
# L_BRACE       ::= '{'
# R_BRACE       ::= '}'
# L_BRACKET     ::= '['
# R_BRACKET     ::= ']'
# L_A_BRACKET   ::= '<'
# R_A_BRACKET   ::= '>'
# COMMA         ::= ','
# PIPE          ::= '|'
# HASH          ::= '#'
# BANG          ::= '!'

import mmScan, mmAutomaton, sys

debug = 0

symtable = {}	# the symbol table

def feature(input, stack, level):
	"""FEATURE ::= FEATURE_START | LETTER | FEATURE_END"""
	if debug: print '%sfeature - %s.' % (level * ' ', `input`)
	if not input: raise SyntaxError, "Can't parse Feature in empty input."
	input_backup = input[:]
	type, value = input.pop(0)
	transition = []
	if type == 'L_A_BRACKET':
		# feature_start
		lap = mmAutomaton.Feature_transition(stack, value)
		transition += [lap]
		type, value = input.pop(0)
		if type != 'LETTER':
			raise SyntaxError, 'expected LETTER in %s' % `input`
		lt = mmAutomaton.Feature_transition(stack, value)
		transition += [lt]
		# look for an ID setter
		if input != [] and input[0][0] == 'HASH':
			input.pop(0)	# discard hash sign ('#')
			type, value = input.pop(0)
			if type not in ('LETTER', 'DIGIT'):
				print "Type: ", type, ", value: ", value
				raise SyntaxError, "Letter or digit expected in %s" % `input`
			symtable[value] = ""
			ids = mmAutomaton.IDsetter_transition(stack, value, symtable)
			transition += [ids]
		return mmAutomaton.Transition(stack, transition)
	elif type == 'LETTER':
		lt = mmAutomaton.Feature_transition(stack, value)
		transition += [lt]
		if input != []:
			if input[0][0] == 'R_A_BRACKET':
				# feature_end
				type, value = input.pop(0)
				rap = mmAutomaton.Feature_transition(stack, value)
				transition += [rap]
				# look for an ID checker
				if input != [] and input[0][0] == 'HASH':
					input.pop(0)	# discard hash sign ('#')
					type, value = input.pop(0)
					if type not in ('LETTER', 'DIGIT'):
						raise SyntaxError, "Letter expected in %s" % `input`
					idc = mmAutomaton.IDchecker_transition(stack, value,
						symtable)
					transition += [idc]
				return mmAutomaton.Transition(stack, transition)
		# when we have no '<' or '>', it means a whole feature, in
		# that case we return the two ends.
		lap = mmAutomaton.Feature_transition(stack, '<')
		lt1 = mmAutomaton.Feature_transition(stack, value)
		lt2 = mmAutomaton.Feature_transition(stack, value)
		rap = mmAutomaton.Feature_transition(stack, '>')
		return mmAutomaton.Transition(stack, [lap, lt1, lt2, rap])
	else:
		raise SyntaxError, 'Could not parse feature in %s' % `input_backup`

def spacer(input, stack, level):
	"""SPACER ::= DIGIT COMMA DIGIT"""
	if debug: print '%sspacer - %s.' % (level * ' ', `input`)
	input_backup = input[:]	# make a copy in case of error
	min = 0
	max = sys.maxint
	try:
		if input[0][0] == 'DIGIT':	# min is specified
			min = int(input[0][1])
			input.pop(0)
			if input[0][0] == 'COMMA':
				input.pop(0)
				if input != []:
					if input[0][0] == 'DIGIT':	# max is specified
						max = int(input[0][1])
						input.pop(0)
			else:
				raise SyntaxError, 'Expected COMMA at %s' % input_backup
		elif input[0][0] == 'COMMA':
			input.pop(0)
			if input != []:
				if input[0][0] == 'DIGIT':	# max is specified
					max = int(input[0][1])
					input.pop(0)
		else:
			raise SyntaxError, 'Expected COMMA or DIGIT at %s' % input_backup
	except IndexError, e:
		raise SyntaxError, 'Incomplete spacer in %s' % input_backup
	# create a match transition for the spacer
	return mmAutomaton.Spacer_transition(stack, 'st', min, max)

def range(input, stack, level):
	"""RANGE ::= L_BRACE DIGIT? COMMA DIGIT? R_BRACE"""
	if debug: print '%srange - %s.' % (level * ' ', `input`)
	# min = -sys.maxint-1
	min = 0
	max = sys.maxint
	input_backup = input[:]
	try:
		if input[0][0] != 'L_BRACE':
			raise SyntaxError, 'expected L_PAREN at %s' % input
		input.pop(0)
		# get minimum
		type, value = input[0]
		if type == 'DIGIT':
			# mimimum was specified
			min = value
			input.pop(0)
			# now get comma
			if input[0][0] != 'COMMA':
				raise SyntaxError, 'expected COMMA at %s' % input
			input.pop(0)	# eat comma
			# if there's a number now, that's the max
			type, value = input[0]
			if type == 'DIGIT':
				max = value
				input.pop(0)
		elif type == 'COMMA':
			# minimum not specified (it's ok)
			input.pop(0)
			# if there's a number now, that's the max
			type, value = input[0]
			if type == 'DIGIT':
				max = value
				input.pop(0)
		else:
			# hmm, should have been either a DIGIT or nothing
			raise SyntaxError, 'first term of range is either a DIGIT \
				or empty. (%s)' % input
		# now we should find a '}'
		if input[0][0] != 'R_BRACE':
			raise SyntaxError, 'expected L_PAREN at %s' % input
		input.pop(0)
	except IndexError, e:
		raise SyntaxError, 'range: input is incomplete - %s' % \
			input_backup
	return min, max

def feature_block(input, stack, level):
	"""FEATURE_BLOCK ::= FEATURE+"""
	if debug: print '%sfeature_block - %s.' % (level * ' ', `input`)
	fb = []	
	fb.append(feature(input, stack, level+1))
	if input:
		# Features start with '<' or a letter
		if input[0][0] in ('L_A_BRACKET', 'LETTER'):	
			fb.append(feature(input, stack, level+1))
	return mmAutomaton.Transition(stack, fb)
		
def feature_list(input, stack, level):
	"""FEATURE_LIST ::= FEATURE_BLOCK { SPACER FEATURE_BLOCK } *"""
	if debug: print '%sfeature_list - %s.' % (level * ' ', `input`)
	fl = []
	fl.append(feature_block(input, stack, level+1))	# at least one feature block
	while input:
		# spacers begin with [1-9,]
		if input[0][0] in ('COMMA', 'DIGIT'):
			fl.append(spacer(input, stack, level+1))
			fl.append(feature_block(input, stack, level+1))
		else:
			break
	return mmAutomaton.Transition(stack, fl)

def sequence(input, stack, level):
	"""SEQUENCE :== (FEATURE_LIST | GROUP)+"""
	if debug: print '%ssequence - %s.' % (level * ' ', `input`)
	trn_list = []	# the transitions to join
	if input[0][0] == 'L_PAREN':	# look-ahead for '('
		grp = group(input, stack, level+1)
		trn_list.append(grp)
	else:
		fl = feature_list(input, stack, level+1)
		trn_list.append(fl)
	while input:
		if input[0][0] in ('PIPE', 'R_PAREN', 'L_BRACKET'):
			return mmAutomaton.Transition(stack, trn_list)
		else:
			seq = sequence(input, stack, level)
			trn_list.append(seq)
	return mmAutomaton.Transition(stack, trn_list)

def group(input, stack, level):
	"""GROUP ::= '(' SEQUENCE { '|' SEQUENCE } * ')' { RANGE }"""
	# must start with a '(', let's check
	if debug: print '%sgroup - %s.' % (level * ' ', `input`)
	branch_list = []	# holds the various branches before they're joined
	if input[0][0] != 'L_PAREN':
		raise SyntaxError, "Expected '(' in %s" % `input`
	input.pop(0)
	seq = sequence(input, stack, level+1)
	branch_list.append(seq)
	# continue parsing sequences as long as we see a '|'
	while input[0][0] == 'PIPE':
		input.pop(0)
		seq = sequence(input, stack, level+1)
		branch_list.append(seq)
	# no more pipes, so there must be a ')':
	if input[0][0] != 'R_PAREN':
		raise SyntaxError, "Expected ')' in %s" % `input`
	input.pop(0)
	alt = mmAutomaton.Alternation(stack, 'alt', branch_list)
	# now see if there's a range:
	if input != []:	# might be empty
		if input[0][0] == 'L_BRACE':
			min, max = range(input, stack, level)
			return mmAutomaton.Counter(stack, 'cnt', alt, min, max)
	return alt

def fork(input, stack, level):
	"""FORK ::= '(' FEATURE_LIST { '|' FEATURE_LIST } * ')'"""
	# must start with a '[', let's check
	if debug: print '%sfork - %s.' % (level * ' ', `input`)
	type = 'or'
	branch_list = []	# holds the various branches before they're joined
	if input[0][0] != 'L_BRACKET':
		raise SyntaxError, "Expected '[' in %s" % `input`
	input.pop(0)
	if input[0][0] == 'BANG':
		type = 'and'
		input.pop(0)
	ftl = feature_list(input, stack, level+1)
	branch_list.append(ftl)
	# continue parsing sequences as long as we see a '|'
	while input[0][0] == 'PIPE':
		input.pop(0)
		ftl = feature_list(input, stack, level+1)
		branch_list.append(ftl)
	# no more pipes, so there must be a ')':
	if input[0][0] != 'R_BRACKET':
		raise SyntaxError, "Expected ']' in %s" % `input`
	input.pop(0)
	return mmAutomaton.Fork(stack, 'frk', branch_list, type)

def metamotif(input, stack, level):
	"""METAPROFILE ::= (SEQUENCE|FORK)+"""
	if debug: print '%smetamotif - %s.' % (level * ' ', `input`)
	list = []
	# now handle the core metamotif
	while input != []:
		if input[0][0] == 'L_BRACKET':
			frk = fork(input, stack, level+1)
			list.append(frk)
		else:
			seq = sequence(input, stack, level+1)
			list.append(seq)
	return mmAutomaton.Transition(stack, list)
