################################################################
#            The mmsearch Metamotif Search Engine
#
# Thomas Junier, April 2001
#
# Please see file COPYRIGHT.
#
# The individual states of a metamotif automaton.

# States have two basic attributes: a _status_ (either 'reject' or
# 'accept') that tells  whether the input string can end here; and a
# (possibly empty) _next state_ that may or  may not depend on the next
# input character.

import sys, string, re

class State:
	"""A state class with basic functionality."""
	def __init__(self, label, symbol=None, status='reject'):
		self.label = label		# to identify this state
		self.symbol = symbol
		self.status = status	# 'reject' or 'accept'
		self.next_state = None
	def __str__(self):
		self_string = `self`
		return "%s (%s)" % (self_string, self.label)
	def get_status(self):
		return self.status
	def set_next_state(self, state):
		"""Set the next state."""
		self.next_state = state
	def get_next_state(self, input_string):
		"""Return the next state, either unconditionally or depending
		on the next input_string symbol read."""
		if self.next_state == None:
			return None	# no next state at all.
		elif self.symbol == None:
			return self.next_state	# epsilon transition
		else:
			# match transition, compare `self.symbol' with input_string
			try:
				symbol = input_string.pop(0) 	
			except IndexError, e:	# `input_string' might be empty
				return None
			else:
				if symbol == self.symbol:
					return self.next_state	# symbols match: proceed.
				else:
					return None
	def set_symbol(self, symbol):
		self.symbol = symbol
	def get_symbol(self):
		return self.symbol
	def print_chain(self):
		state = self
		while state:
			print state.label,
			state = state.next_state
		print

class Feature(State):
	"""A Feature object ignores spacers (if any). If the spacer's
	length is important, use a Spacer object to check it."""
	def get_next_state(self, input_string):
		try:
			if input_string[0] in string.digits or \
				input_string[0] in ('-', ' ', '#'):	# ignore digits, dash, space, and hash
				input_string.pop(0)
				return self
			else:
				return State.get_next_state(self, input_string)
		except IndexError, e:
			return None

class Any(State):
	"""Accepts any value (like the 'x' in Prosite patterns, or '.' in Perl-style regular
	expressions. Consumes the input character."""
	def get_next_state(self, input_string):
		"""Return the next state, consuming the input symbol."""
		try:
			input_string.pop(0)
		except IndexError, e:
			return None
		return self.next_state

class IDsetter(State):
	"""Sets a variable (whose name is passed as param 'name')
	in the symbol table to the value of a feature ID as found in
	the input string. An IDsetter can never be an acceptor state,
	but it unconditionally transits to the next state. This state
	is used with an IDchecker instance."""
	def __init__(self, label, name, symtable):
		State.__init__(self, label) # status is 'reject' by default, which is ok
		self.name = name
		self.symtable = symtable
	def get_next_state(self, input_string):
		"""Before returning the next state (probably a Feature
		or Spacer), this method sets the `name' variable to its
		appropriate value, which is a number unique to a particular
		motif an thus can be used to distinguish two different motifs
		of the same kind (e.g., two disulfide bridges)."""
		value = ""
		try:
			c = input_string.pop(0)
			if c != '#': raise SyntaxError, \
				"IDsetter: expected a '#' before %s" % `input_string`
			while 1:	
				c = input_string.pop(0)
				if c in string.digits:
					value += c	# digit: append to value
				else:
					# non-digit (should be a dash...)
					self.symtable[self.name] = value 
					return self.next_state
		except IndexError, e:
			return None

class IDchecker(State):
	"""The companion of IDsetter. This checks that the feature ID it
	finds in the input string is the same as that stored in the symbol
	table under `name'. If not, returns None (and rejects input)."""
	def __init__(self, label, name, symtable):
		State.__init__(self, label)
		self.name = name
		self.symtable = symtable
	def get_status(self):
		if self.ftID == self.symtable[self.name]:
			return 'accept'
		else:
			return 'reject'
	def reset(self):
		self.symtable[self.name] = ""
	def get_next_state(self, input_string):
		value = ""
		try:
			c = input_string.pop(0)
			if c != '#': raise SyntaxError, \
				"IDSetter: expected a '#' before ", `input_string`
			while 1:	
				c = input_string.pop(0)
				if c in string.digits:
					value += c	# digit: append to value
				else:
					self.ftID = value # non-digit (should be a dash...)
					if self.get_status() == 'accept':
						self.reset()
						return self.next_state
					else:
						return None
		except IndexError, e:
			return None
		

class Spacer(State):
	"""Spacers have special behaviour: they accept any string
	of digits, '-' or ' ',	which represents the positions of
	consecutive features (or feature ends). Upon leaving the state,
	the difference between these two numbers is compared to the
	state's two parameters `min' and `max', and the transition is
	allowed only if min <= #residues <= max."""
	def __init__(self, label, min=0, max=sys.maxint, status='reject'):
		State.__init__(self, label, status)
		self.min = min
		self.max = max
		self.digits = ""	# will evtl. hold all digits of spacer length
		self.regex = re.compile(r'\d+')
	def reset(self):
		self.digits = ""	# clear digits
	def get_status(self):
		if self.digits == "":
			return 'reject'
		numbers = self.regex.findall(self.digits)
		# Check that we have both positions - this may not be the case if one
		# motif is absent and would have been the first or last had it been
		# present.
		if len(numbers) < 2:
			return 'reject'
		# We take the two LAST numbers, as there may be three (when there is no
		# IDchecker before this Spacer), in which case the first one is just the
		# feature's ID.
		length = int(numbers[-1]) - int(numbers[-2]) -1
		if length >= self.min and length <= self.max:
			self.status = 'accept'
		else:
			self.status = 'reject'
		return self.status
	def set_next_state(self, state):
		self.next_state = state
	def get_symbol(self):
		return None	# epsilon
	def get_next_state(self, input_string):
		try:	# input_string can be empty
			symbol = input_string.pop(0)
		except IndexError, e:
			# input_string is empty
			if self.get_status() == 'accept':
				self.reset()
				return self.next_state
			else:
				self.reset()
				return None
		else:
			# input_string isn't empty
			if symbol in string.digits or symbol in (' ', '-', '#'):
				# it's a digit, space, or dash: remember
				# it, and loop back to self
				self.digits = self.digits + symbol
				return self
			else:
				# not a digit, space, or dash: push it
				# back and check for transition
				input_string.insert(0, symbol)	# push back symbol
				# perform test
				if self.get_status() == 'accept':
					# ok, number is between min and max
					self.reset()
					return self.next_state
				else:
					self.reset()
					return None

class Alternator_Head(State):
	"""Handles alternation, in conjunction with an Alternator_Tail
	object. The state maintains a list of next states. Each time
	get_next_state() is called, a next state is popped from the list.
	When the list is empty, all the branches of the alterntive have
	been tried."""
	def __init__(self, stack, label):
		State.__init__(self, label)
		self.next_states = []	# the list of next states
		self.input_string = None		# copy of the input_string, for backtracking
		self.stack = stack		# a stack for backtracking
	def reset(self):
		# restore the original list of states
		self.next_states = self.next_states_backup[:]	
		self.input_string = None
		# print "Alt: next states after reset(): %s" % \
		#	`self.next_states`
	def set_next_state(self, state):
		self.next_states.append(state)
	def get_next_state(self, input_string):
		if self.input_string == None:	# input_string has not yet been stored
			self.input_string = input_string[:]	# a COPY (not just a ref)!
			self.next_states_backup = self.next_states[:] # see reset()
			# if __debug__: print 'alt: pushing self onto stack' 
			self.stack.append(self)	# push self onto stack
		# each branch starts with a fresh copy
		# if __debug__: print "alt %s: possible branches are %s" % (self.label,
		#	`self.next_states`)
		while len(input_string) > 0:
			input_string.pop(0)
		input_string.extend(self.input_string[:])
		if len(self.next_states) > 0:
			return self.next_states.pop(0)
		else:
			# remove self from the stack and reset
			self.stack.pop(-1)
			self.reset()
			return None

class Alternator_Tail(State):
	"""Works in conjunction with an Alternator_Head. The task of this
	state is to reset the alternator when a branch matches."""
	def __init__(self, stack, label):
		State.__init__(self, label)
		self.stack = stack
		self.status = 'accept'
	def set_head(self, head):
		self.head = head
	def get_next_state(self, input_string):
		self.stack.pop(-1)	
		self.head.reset()	# ready for next time
		return State.get_next_state(self, input_string)

class Counter_Head(State):
	"""Counts the number of times it is accessed. Works in conjunction
	with a Counter_Tail object."""
	def __init__(self, stack, label):
		State.__init__(self, label)
		self.stack = stack
		self.loop_count = 0
		self.tail = None	# the "tail" state
	def set_tail(self, tail):
		self.tail = tail
	def reset(self):
		self.loop_count = 0
	def get_next_state(self, input_string):
		if self.loop_count == 0:
			# first time...
			self.stack.append(self.tail)	# fallback on tail if blocked
			# if __debug__: print "First time in %s" % self.label
			self.loop_count = self.loop_count + 1
		else:
			self.loop_count = self.loop_count + 1
			# if __debug__: print "%s accessed %d time(s)" % (self.label,
			#	self.loop_count)
		return self.next_state	# always try another loop
	def get_loop_count(self):
		return self.loop_count

class Counter_Tail(State):
	"""Works in conjunction with a Counter_Head to implement a
	counter."""
	def __init__(self, stack, label, min, max):
		State.__init__(self, label)
		self.min = int(min)
		self.max = int(max)
		self.stack = stack
	def set_head(self, head):
		self.head = head
	def get_next_state(self, input_string):
		self.stack.pop(-1)	# pop oneself off the stack
		loop_count = self.head.get_loop_count()
		# the last loop always fails, so we count one less
		loop_count = loop_count - 1
		# print 'CT: min = %s, max = %s, count = %d' % (self.min, self.max,
		#	loop_count)
		if loop_count >= self.min and loop_count <= self.max:
			self.status = 'accept'
			self.head.reset()
			return State.get_next_state(self, input_string)
		else:
			self.status = 'reject'
			self.head.reset()
			return None

class Fork_head(State):
	"""Implements a parallel alternator, or fork. Nothing to do with
	transcription factors :-). [Note: there used to be a companion
	Fort_tail class, but there isn't anymore. The name has been kept
	to avoid breaking things.]"""
	def __init__(self, stack, label, type):
		State.__init__(self, label)
		self.stack = stack
		self.type = type
		self.input_string_backup = None
	def set_branches(self, state_list):
		"""Set the list of next states."""
		# print 'FH.state_list:', `state_list`
		self.state_list = state_list
		self.state_list_backup = state_list[:]
	def reset(self):
		self.state_list = self.state_list_backup[:]
	def get_next_state(self, input_string):
		"""Overridden to allow sequential testing of many
		simultaneously occurring branches. Does a bit what the
		basic Automaton does."""
		debug = 0	# switch to 1 when debugging
		# for each branch, try to match and store remaining input.
		if debug: print "Entering fork...", self
		if debug: print "Input   :", string.join(input_string, '')
		if debug: print "Branches:", self.state_list
		branch_results = []
		immediate_branches = 0	# must be at least 1 for Fork to match
		for branch in self.state_list:
			if debug: print "\nNew branch"
			branch_data = self.try_branch(branch, input_string[:])
			if debug: print branch_data
			# discard results which did not match
			if branch_data:
				immediate_match = branch_data.pop()
				if debug: print 'Immediate:', immediate_match
				immediate_branches += immediate_match
				branch_results.append(branch_data)
				if debug: self.shift_print(branch_data[0], branch_data[1])
		# return None if no branch matches
		if branch_results == []: return None 
		# return None if not even one branch matched from the first
		# position of input
		if immediate_branches == 0: return None
		# get overlapping region, according to type
		if self.type == 'and':
			if len(branch_results) < len(self.state_list): return None
			input_after_match = self.skip_total_overlap(branch_results)
			if input_after_match == None: return None
		elif self.type == 'or':
			input_after_match = self.skip_longest_overlap(branch_results)
		else:
			raise "Unknown Type of Fork_head"
		# empty input string and set it to input past match
		while input_string != []:
			input_string.pop(0)
		input_string.extend(input_after_match)
		if debug: print 'input (end of fork):', input_string
		return self.next_state
	def try_branch(self, branch, input_string):
		"""Try one branch of the fork. Returns a list composed of the
		input form the start of the match (first matching feature end,
		other chars don't count); the input from the first position after
		the match; and whether the branch matched from the start (IOW,
		whether there were shifts before the first match or
		transition). The first two elements of the return list
		can then be used by get_next_state() to find overlapping
		branches. Whether a branch matches from the start is
		important, as there must be at least one such branch in a Fork."""
		debug = 0	# switch to 1 when debugging
		current_state = branch
		length = len(input_string)
		match_start = ""# input string where match starts
		transited = 0	# did we at least do 1 feature (<[a-Z]>) transition?
		immediate_match = 1	# did the match start at the first position
		self.prev_char	= ''
		self.prev_label	= ''
		while 1:
			if debug: print "current:", current_state
			if debug: print "input  :", string.join(input_string, '')
			try:
				current_char = input_string[0]
				next_state = current_state.get_next_state(input_string)
			except IndexError, e:	# hit the end of input_string
				return []
			if next_state == None:	# input_string leads nowhere
				if current_state.get_status() == 'accept':
					# ok, this branch matches...
					if debug: print 'accept'
					return [match_start, input_string, immediate_match]
				elif input_string!= []:
					if debug: print 'blocked on', current_char
					if transited == 0:
						if debug: print 'Not yet transited - differed match.'
						immediate_match = 0
					if debug: print 'blocked - popping.'
					if len(input_string) == length:
						# input_string was not consumed: discard present
						# state, so we shift to the next one.
						input_string.pop(0)
						# -> next while: 
				else:
					# no more input, and no accept: return empty
					# string.
					return []
			else:
				if debug: print 'transition ->', next_state
				# check that the next char also matches
				if self.check_chars(current_state, current_char,
					input_string[:]):
					transited = 1
					# record start of match if not already done
					if match_start == "":
						match_start = [current_char] + input_string
				else:
					# if not yet transited, rewind branch
					if transited == 0:
						next_state = branch	
				# update state
				current_state = next_state
	def check_chars(self, cur_state, cur_char, input):
		"""Check that this and the next characters correspond to this
		and the next state's labels."""
		debug = 0
		cur_label = cur_state.label
		next_label = cur_state.next_state.label
		if not input: return 0
		next_char = input[0]
		if debug: print 'cur_char: %s, cur_label: %s; \
			next_char: %s, next_label: %s' \
			% (cur_char, cur_label, next_char, next_label)
		return (cur_char in string.letters or cur_char in ('<', '>')) and \
			cur_char == cur_label and next_char == next_label

	def shift_print(self, list1, list2):
		"""A debugging method that takes two lists, supposed to have a
		common suffix, and repends '.'s to the shorter one. This is
		useful for displaying input before and after a match (the
		difference being the matched region itself)."""
		s1 = string.join(list1, '')
		s2 = string.join(list2, '')
		l1 = len(s1)
		l2 = len(s2)
		if l1 == l2:
			print s1
			print s2
		else:
			if l1 < l2:
				short, long = s1, s2
			else:
				short, long = s2, s1
			print long
			print '%s%s' % ((len(long)-len(short))*'.', short)
	def skip_longest_overlap(self, branch_results):
		""" Find the longest set of overlapping branches, and return
		remaining input after match. Default is first branch, and
		will be extended only if overlapping branches are found."""
		# sort the matching branches by increasing position of match
		# start (IOW, by _decreasing_ length of
		# input-after-match-start).
		branch_results.sort(lambda y, x: len(x[0]) - len(y[0]))
		input_after_match = branch_results[0][1]	# first branch
		for branch in branch_results:
			# check overlap: branch[0] is input after start, branch[1]
			# input after end of match. Inclusions, however, are
			# rejected
			if len(branch[0]) > len(input_after_match) and \
			len(branch[1]) < len(input_after_match):
				input_after_match = branch[1]
		return input_after_match
	def skip_total_overlap(self, branch_results):
		""" Find the longest set of overlapping branches, and return
		remaining input after match. Default is first branch, and
		will be extended only if overlapping branches are found. All
		branches must match in a contiguous region."""
		# sort the matching branches by increasing position of match
		# start (IOW, by _decreasing_ length of
		# input-after-match-start).
		debug = 0
		if debug: print
		if debug: print len(branch_results), 'branches.'
		branch_results.sort(lambda y, x: len(x[0]) - len(y[0]))
		if debug:
			for branch in branch_results:
				self.shift_print(branch[0], branch[1])
		input_after_match = branch_results[0][1]	# first branch
		count = 1	# we start from second branch
		for branch in branch_results[1:]:
			if debug: print 'from match start ', string.join(branch[0], '')
			if debug: print 'input after match', \
				string.join(input_after_match, '')
			# check overlap: branch[0] is input after start, branch[1]
			# input after end of match. Inclusions, however, are
			# rejected
			if len(branch[0]) > len(input_after_match):
				if debug: print '-> overlap or inclusion'
				# overlap or inclusion
				count += 1
				if debug: print '\tfrom match start ', string.join(branch[1], '')
				if debug: print '\tinput after match', \
					string.join(input_after_match, '')
				if len(branch[1]) < len(input_after_match):
					if debug: print '\t-> overlap s. s.'
					# overlap sensu stricto: update input_after_match
					input_after_match = branch[1]
		if count < len(branch_results): return None
		return input_after_match
